import pytest
import requests
import requests_mock
from unittest.mock import patch

from pubtools.sign.clients.registry import ContainerRegistryClient


@pytest.fixture
def mocked_auth_config_files():
    with patch("pubtools.sign.clients.registry.open") as mock_open:
        with patch("pubtools.sign.clients.registry.os.path.exists") as mock_exists:
            mock_exists.return_value = True
            with patch("pubtools.sign.clients.registry.json") as mocked_json:
                mocked_json.load.return_value = {
                    "auths": {"registry.example.com": {"auth": "dXNlcm5hbWU6cGFzc3dvcmQ="}}
                }
                yield mock_open, mocked_json


def test_resolve_authentication(mocked_auth_config_files):
    client = ContainerRegistryClient()
    auth = client.resolve_authentication("registry.example.com/foo/bar:latest")
    assert auth == ("username", "password")


def test_authenticate_to_registry(mocked_auth_config_files):
    client = ContainerRegistryClient()
    with requests_mock.Mocker() as m:
        m.get(
            "https://auth.example.com/auth?service=example.com&scope=repository%3Afoo%2Fbar%3Apull",
            json={"token": "some-token"},
        )
        token = client.authenticate_to_registry(
            "registry.example.com/foo/bar",
            'www-authenticate: Bearer realm="https://auth.example.com/auth",'
            'service="example.com",'
            'scope="repository:foo/bar:pull"',
        )
        assert token == "some-token"


def test_authenticate_to_registry_error(mocked_auth_config_files):
    client = ContainerRegistryClient()
    with requests_mock.Mocker() as m:
        m.get(
            "https://auth.example.com/auth?service=example.com&scope=repository%3Afoo%2Fbar%3Apull",
            status_code=401,
        )
        with pytest.raises(requests.exceptions.HTTPError):
            client.authenticate_to_registry(
                "registry.example.com/foo/bar",
                'www-authenticate: Bearer realm="https://auth.example.com/auth",'
                'service="example.com",'
                'scope="repository:foo/bar:pull"',
            )



from pubtools.sign.signers.msgsigner import (
    ContainerSignResult,
    ClearSignResult,
)


def test_containeroperation_result_to_dict():
    assert ContainerSignResult(
        results=["test"], signing_key="signing_key", failed=False
    ).to_dict() == {
        "results": ["test"],
        "signing_key": "signing_key",
        "failed": False,
    }


def test_clearoperation_result_to_dict():
    assert ClearSignResult(outputs=["test"], signing_key="signing_key").to_dict() == {
        "outputs": ["test"],
        "signing_key": "signing_key",
    }



import collections
import json
import socket
import logging
from multiprocessing import Process
import threading
import tempfile
import uuid
import os
import sys

from unittest.mock import patch

from .conftest_msgsig import (  # noqa: F401
    f_msg_signer,  # noqa: F401
    f_config_msg_signer_ok,  # noqa: F401
    f_config_msg_signer_ok2,  # noqa: F401
    f_config_msg_signer_aliases,  # noqa: F401
)  # noqa: F401
from .conftest_cosignsig import (  # noqa: F401
    f_cosign_signer,  # noqa: F401
    f_config_cosign_signer_ok,  # noqa: F401
    f_config_cosign_signer_aliases,  # noqa: F401
    f_config_cosign_signer_no_auth,
)  # noqa: F401


from proton import Endpoint
from proton.reactor import Container
import proton
from proton import Message

from pytest import fixture

from pubtools.sign.clients.msg import _MsgClient


LOG = logging.getLogger("pubtools.sign.signers.radas")
LOG.addHandler(logging.StreamHandler(sys.stdout))


class _Queue(object):
    def __init__(self, address, dynamic=False):
        self.address = address
        self.dynamic = dynamic
        self.queue = collections.deque()
        self.consumers = []

    def subscribe(self, consumer):
        LOG.debug("QUEUE SUBSCRIBED: %s %s", self.address, consumer)
        self.consumers.append(consumer)

    def unsubscribe(self, consumer):
        LOG.debug("QUEUE UNSUBSCRIBED: %s %s", self.address, consumer)
        if consumer in self.consumers:
            self.consumers.remove(consumer)
        return len(self.consumers) == 0 and (self.dynamic or len(self.queue) == 0)

    def publish(self, message):
        self.queue.append(message)
        self.dispatch()

    def dispatch(self, consumer=None):
        if consumer:
            c = [consumer]
        else:
            c = self.consumers
        LOG.debug("QUEUE DISPATCH TO %s %s", self.address, c)
        while self._deliver_to(c):
            pass

    def _deliver_to(self, consumers):
        try:
            result = False
            for c in consumers:
                if c.credit:
                    message = self.queue.popleft()
                    LOG.debug("QUEUE DELIVER TO %s %s", self.address, c, message)
                    c.send(message)
                    result = True
            return result
        except IndexError:  # no more messages
            LOG.debug("QUEUE NOTHING TO DELIVER %s", self.address)
            return False


class _Broker(_MsgClient):
    def __init__(self, url):
        super().__init__(errors=[])
        self.url = url
        self.queues = {}

    def on_start(self, event):
        print("BROKER on start", self.url)
        LOG.info("BROKER on start", self.url)
        self.acceptor = event.container.listen(self.url)

    def _queue(self, address):
        if address not in self.queues:
            self.queues[address] = _Queue(address)
        return self.queues[address]

    def on_link_opening(self, event):
        LOG.info(
            "BROKER on_link_opening event",
            event.link,
            "source addr:",
            event.link.source.address,
            "remote source addr",
            event.link.remote_source.address,
            "target addr:",
            event.link.target.address,
            "remote target addr",
            event.link.remote_target.address,
        )
        print(
            "BROKER on_link_opening event",
            event.link,
            "source addr:",
            event.link.source.address,
            "remote source addr",
            event.link.remote_source.address,
            "target addr:",
            event.link.target.address,
            "remote target addr",
            event.link.remote_target.address,
        )
        if event.link.is_sender:
            if event.link.remote_source.dynamic:
                address = str(uuid.uuid4())
                event.link.source.address = address
                q = _Queue(address, True)
                self.queues[address] = q
                q.subscribe(event.link)
            elif event.link.remote_source.address:
                event.link.source.address = event.link.remote_source.address
                self._queue(event.link.source.address).subscribe(event.link)
        elif event.link.remote_target.address:
            event.link.target.address = event.link.remote_target.address

    def _unsubscribe(self, link):
        if link.source.address in self.queues and self.queues[link.source.address].unsubscribe(
            link
        ):
            del self.queues[link.source.address]

    def on_link_closing(self, event):
        LOG.info(">> BROKER On link closing", event)
        if event.link.is_sender:
            self._unsubscribe(event.link)

    def on_disconnected(self, event):
        LOG.info(">> BROKER On disconnected", event)
        self.remove_stale_consumers(event.connection)

    def remove_stale_consumers(self, connection):
        link = connection.link_head(Endpoint.REMOTE_ACTIVE)
        LOG.info("BROKER removing stale consumer", link)
        while link:
            if link.is_sender:
                self._unsubscribe(link)
            link = link.next(Endpoint.REMOTE_ACTIVE)

    def on_sendable(self, event):
        LOG.info("BROKER on_sendable", event.link.source.address)
        self._queue(event.link.source.address).dispatch(event.link)

    def on_message(self, event):
        LOG.info("BROKER ON MESSAGE", event.message)
        address = event.link.target.address
        if address is None:
            address = event.message.address
        LOG.debug("BROKER publish", address)
        self._queue(address).publish(event.message)


class _BrokenBroker(_Broker):
    def on_sendable(self, event):
        LOG.info("BROKER on_sendable", event.link.source.address)
        self._queue(event.link.source.address).dispatch(event.link)
        raise ValueError("Simulated broker error")
        event.on_link_error(event)


class _FakeMsgSigner(proton.handlers.MessagingHandler):
    def __init__(self, broker_urls, listen_to, send_to, cert, ca_cert, received_messages):
        self.broker_urls = broker_urls
        self.listen_to = listen_to
        self.send_to = send_to
        self.ssl_domain = proton.SSLDomain(proton.SSLDomain.MODE_CLIENT)
        self.ssl_domain.set_peer_authentication(proton.SSLDomain.ANONYMOUS_PEER)
        self.received_messages = received_messages
        super().__init__()
        self.to_send = []

    def on_start(self, event):
        conn = event.container.connect(
            urls=self.broker_urls, ssl_domain=self.ssl_domain, sasl_enabled=False
        )
        self.receiver = event.container.create_receiver(conn, self.listen_to)

    def on_message(self, event):
        LOG.debug("RADAS on message", event.message)
        headers = event.message.properties
        self.received_messages.append(event.message)

        if headers.get("mtype") == "container_signature":
            reply_message = ""
        if headers.get("mtype") == "clearsig_signature":
            reply_message = ""
        else:
            reply_message = event.message.body

        reply = Message()
        reply.address = self.send_to
        LOG.debug("Send to", self.send_to)
        reply.body = reply_message
        reply.properties = event.message.properties

        sender = event.container.create_sender(event.connection)
        LOG.debug("RADAS Sending", reply)
        sender.send(reply)
        LOG.debug("RADAS Sent")

    def on_sendable(self, event):
        LOG.debug("RADAS on sendable")
        if not self.to_send:
            LOG.debug("RADAS Nothing to send")
            return
        message_to_send = self.to_send.pop(0)
        event.sender.send(message_to_send)


class _StrayFakeMsgSigner(_FakeMsgSigner):
    def on_message(self, event):
        headers = event.message.properties
        self.received_messages.append(event.message)

        if headers.get("mtype") == "container_signature":
            reply_message = ""
        if headers.get("mtype") == "clearsig_signature":
            reply_message = ""
        else:
            reply_message = json.loads(event.message.body)

        reply = Message()
        reply.address = self.send_to
        reply_message["msg"]["request_id"] += "1"
        reply.body = json.dumps(reply_message)

        reply.properties = event.message.properties

        sender = event.container.create_sender(event.connection)
        sender.send(reply)


def run_broker(broker, stdout):
    sys.stdout = stdout
    broker.run()


@fixture(scope="session")
def f_msgsigner_listen_to_topic():
    return "topic://Topic.pubtools.sign"


@fixture(scope="session")
def f_msgsigner_send_to_queue():
    return "topic://Topic.signatory.sign"


@fixture(scope="session")
def f_msgsigner_listen_to_topic_stray():
    return "topic://Topic.pubtools.sign.stray"


@fixture(scope="session")
def f_msgsigner_send_to_queue_stray():
    return "topic://Topic.signatory.sign.stray"


@fixture(scope="session")
def f_received_messages():
    return []


@fixture
def f_cleanup_msgsigner_messages(f_received_messages):
    return f_received_messages.clear()


@fixture(scope="session")
def f_find_available_port():
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


@fixture(scope="session")
def f_find_available_port_for_broken():
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


@fixture(scope="session")
def f_qpid_broker(f_find_available_port):
    LOG.info("starting broker", f"localhost:{f_find_available_port}")
    broker = Container(_Broker(f"localhost:{f_find_available_port}"))
    p = Process(target=run_broker, args=(broker, sys.stdout))
    p.start()
    yield (broker, f_find_available_port)
    LOG.info("destroying qpid broker")
    p.terminate()


@fixture(scope="session")
def f_broken_qpid_broker(f_find_available_port_for_broken):
    LOG.debug("starting broker", f"localhost:{f_find_available_port_for_broken}")
    broker = Container(_BrokenBroker(f"localhost:{f_find_available_port_for_broken}"))
    p = Process(target=broker.run, args=())
    p.start()
    yield (broker, f_find_available_port_for_broken)
    LOG.debug("destroying qpid broker")
    p.terminate()


@fixture(scope="session")
def f_fake_msgsigner(
    f_find_available_port,
    f_msgsigner_listen_to_topic,
    f_msgsigner_send_to_queue,
    f_received_messages,
):
    fr = _FakeMsgSigner(
        [f"localhost:{f_find_available_port}"],
        f_msgsigner_listen_to_topic,
        f_msgsigner_send_to_queue,
        "",
        "",
        f_received_messages,
    )
    frc = Container(fr)
    t = threading.Thread(target=frc.run, args=())
    t.start()
    yield fr, frc, f_received_messages
    frc.stop()


@fixture(scope="session")
def f_fake_msgsigner_stray(
    f_find_available_port,
    f_msgsigner_listen_to_topic_stray,
    f_msgsigner_send_to_queue_stray,
    f_received_messages,
):
    fr = _StrayFakeMsgSigner(
        [f"localhost:{f_find_available_port}"],
        f_msgsigner_listen_to_topic_stray,
        f_msgsigner_send_to_queue_stray,
        "",
        "",
        f_received_messages,
    )
    frc = Container(fr)
    t = threading.Thread(target=frc.run, args=())
    t.start()
    yield fr, frc, f_received_messages
    frc.stop()


@fixture
def f_client_certificate():
    with tempfile.NamedTemporaryFile() as tmpf:
        tmpf.write(
            """
-----BEGIN PRIVATE KEY-----
MIIEvQIBADANBgkqhkiG9w0BAQEFAASCBKcwggSjAgEAAoIBAQC1y5rshMkYqfP2
k7z2IJXto3AvLCAYDN9WX5mUUFgPFAFur38bDopqj5dSkXlR5f3MwZeRCNsxRXKD
USjv1T9HjD2t2D7eOAQhhx7dtgMPEYRJes02/ejzGojqrXEoG3TsCf8wBmb5ALDc
VrgEWShRRGi3qzsDXijMtzns+30+/rOfR68+353iLXTJ3g/MkVRLks6xwq35ZbH4
Ichu/6ISRU7Lk3tOq811a/SFuHdSLrOzU13VDK+6nQiWMUPbHy6OHK88dC+6mRws
RJuqHg0cKSIA5aiE7T7o9S3v10P0wJnn8B7d+bsHZJPxqLxFdbQ5Ogjrc195TEmf
kAd4kB8BAgMBAAECggEABwvEdhfuK58cLBWVKoCwMOr/jEuAdbAriI9pXlvIEvMI
fG3PvkinrZpBwB7wad/BMMo8za+vP6tQdTWcQaHlK0ispBAUXTRkeBhypj8Rgqgw
FXE9kozgJkp0LazR3L8mLqcTSX0pCOv5ftu7U3tRZkdxi9NWlOKFkCwmkghcBgjE
M2LZx4mCNbtD0WQYqMr0HdGJsedCNRrmCAZ5eoTM+HIh8C2eCRT4mBirNLIHsg/d
TqLI9TN0JvDkpTHt7W94Cxi7th5uLrXfE+D5cbHT0OV4L/kwQj3JjZDekUTCqljX
OcC8X0l8+SYo6+qlc0mmd324LzGmD+fuP9LXnHkc8QKBgQDwUABdZKZWRWMdypj4
jgzSBZa1k6cDt39373MFwhk+c5nYUm7ijsRncz85Vb8n0vwY6K0DWuP3yNfuKyVe
BDXZHfgncz3hc+pBa3B/NIxDmI16D2SEomhe+/JyHh1wnOI5DPJFSrWhYtuMvKpv
QBpQ0rNP8Xq+6c/Gg06g1SFRkQKBgQDBqbCpPOt6+dlR/7aRzbWBe9hUqYegRKw+
Is+Ap9hcup7zstsKOdjBGt89wV7BV9KQQnE69uOqBx6CdqEUId47C3F661hT/npR
B9bBdwW7m8FKbu5i86d1V3ddvdNbIkAeqwjSpnkrLzITPJ74w9yhxAxqO/7wMAs1
PBTpNHo+cQKBgE/f128zYBI2t+4UA+pBlMNN9jzeGdojaKvdm9ajIC7gz5bWN2L4
XxGffbk55fJ/ryk8VR1TXYhjaloQXzgzoA5NZsj+Behk1czuwBKXzbM+BnA2o4tu
S9CeX4RMvC5NBug9hF1BqsM8j4rkvqWBof2ROuZsdgb0wgnSZRUSIiPxAoGAHMUg
wYOTWAmWB2B9tttgg4Pqd2lYBK8vB2wUd4B33A69XmbLs5E0ajubvojjksWBOn0k
ZSYYXEICfk8xTtRZN1xT13bvAEtl0HPhq4wLBfv1kyE3uOuJjR0ZVovEwl0sOWIf
RWwFxCyWu9TdqQcv17hQP9f536TDhX0PfjWVk4ECgYEAp5ze8TFxH8PyTqGOGikn
aKhWL4DD3dNoXCYOrcEmTKOXftlvlrR+6VjCT0DFro9QxQM9S7BEhugmbm/kQoD3
0jxXWkX9wkN2thOrUwJBHBtj/PUqA+s5bbi//V8pSmdSfXn3HgKaXD40sAoX8HWn
8n8nYo3EMM6kk5gbZToN33k=
-----END PRIVATE KEY-----
-----BEGIN CERTIFICATE-----
MIIDdTCCAl2gAwIBAgIUSd6+UGxdN+U6W7tYnZwrY1Tn5NYwDQYJKoZIhvcNAQEL
BQAwSjFIMAkGA1UEBhMCVVMwGQYDVQQDDBJwdWJ0b29scy1zaWduLXRlc3QwIAYK
CZImiZPyLGQBAQwScHVidG9vbHMtc2lnbi10ZXN0MB4XDTI0MDMwNjEyMzY0MloX
DTI0MDQwNTEyMzY0MlowSjFIMAkGA1UEBhMCVVMwGQYDVQQDDBJwdWJ0b29scy1z
aWduLXRlc3QwIAYKCZImiZPyLGQBAQwScHVidG9vbHMtc2lnbi10ZXN0MIIBIjAN
BgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAtcua7ITJGKnz9pO89iCV7aNwLywg
GAzfVl+ZlFBYDxQBbq9/Gw6Kao+XUpF5UeX9zMGXkQjbMUVyg1Eo79U/R4w9rdg+
3jgEIYce3bYDDxGESXrNNv3o8xqI6q1xKBt07An/MAZm+QCw3Fa4BFkoUURot6s7
A14ozLc57Pt9Pv6zn0evPt+d4i10yd4PzJFUS5LOscKt+WWx+CHIbv+iEkVOy5N7
TqvNdWv0hbh3Ui6zs1Nd1Qyvup0IljFD2x8ujhyvPHQvupkcLESbqh4NHCkiAOWo
hO0+6PUt79dD9MCZ5/Ae3fm7B2ST8ai8RXW0OToI63NfeUxJn5AHeJAfAQIDAQAB
o1MwUTAdBgNVHQ4EFgQU8A73Ay/YOz2vLvTPwfptk48pndQwHwYDVR0jBBgwFoAU
8A73Ay/YOz2vLvTPwfptk48pndQwDwYDVR0TAQH/BAUwAwEB/zANBgkqhkiG9w0B
AQsFAAOCAQEAnOmCHWw5UAxJJwxzEW0KRx56qmVawJ6QJCEvBLMXqZuR+kR6Sw7b
WysDTqsD091GVSpvu76RXJLSH/BiP+HbmjPLtb5qb55XM+i7dcz9pIB505jovvC4
t+OA/JZtx/8OmtIhhF1Se2n6gj8dG1H0tuaKbW3E95K7pC59yyW0zKP08ensy2QD
MfggkSRenn4VoA91+jRMylnawn4jIUlWyvUsvturSVz2WP7NmlYTnlCcifyBuYqq
IQ1XIX6F1jjody+3I+8b2tBpaPuNXDAtfoEoUWZW0ToTfAi+6Li7IMjXRZ6wPVxU
aoKJ9jBURYeYzd/Zi2RPLpjt8TYPir8vKQ==
-----END CERTIFICATE-----

""".encode(
                "utf-8"
            )
        )
        tmpf.flush()
        yield tmpf.name


@fixture
def f_ca_certificate():
    with tempfile.NamedTemporaryFile() as tmpf:
        tmpf.write(
            """
-----BEGIN CERTIFICATE-----
MIIDtzCCAp+gAwIBAgIUATUd1WliG6ETZqKP8EZyijG9xUIwDQYJKoZIhvcNAQEL
BQAwazELMAkGA1UEBhMCdVMxFTATBgNVBAcMDERlZmF1bHQgQ2l0eTEQMA4GA1UE
CgwHUmVkIEhhdDEbMBkGA1UECwwSQ2xvdWQgRGlzdHJpYnV0aW9uMRYwFAYDVQQD
DA1wdWJ0b29scy1zaWduMB4XDTIzMDUyNTEyNDQwMloXDTI4MDUyMzEyNDQwMlow
azELMAkGA1UEBhMCdVMxFTATBgNVBAcMDERlZmF1bHQgQ2l0eTEQMA4GA1UECgwH
UmVkIEhhdDEbMBkGA1UECwwSQ2xvdWQgRGlzdHJpYnV0aW9uMRYwFAYDVQQDDA1w
dWJ0b29scy1zaWduMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA2e8m
tgCPFummVobuc+IHdTHfdDVYnMYyp3+ZRrzc5yCYjRpeIoABzcI/aZza2kljHNKZ
yan6LnsdD45xuyhDVmgMHij7Luq0p8ibQAQvmf6kh1pMgFy3Mtsm+lwT99Bt8gNN
YUiahSkN+vZa3eswZTzu5z5RkztzZt4O9qzsdUR7tKPjB5OlvsZFyFnvgtnAByqh
bLpe/YHR/A79TWgzZFBt67/f4ghGHUtN+CPB6e+TLKQ9QRsOqZLhuMwlNsJSBc8k
duvaXYDU3w+GXMci7pWIk3HM2Z9m0AxizZe8ygz/wworSxC8CS2WwPil3W9ft4iT
LTxBoEtN2MQUTMyo9QIDAQABo1MwUTAdBgNVHQ4EFgQUD+IG9eOMO65c5BCgycdw
60t7z3owHwYDVR0jBBgwFoAUD+IG9eOMO65c5BCgycdw60t7z3owDwYDVR0TAQH/
BAUwAwEB/zANBgkqhkiG9w0BAQsFAAOCAQEAsX9O5pnlmmXfm6Vx98bOp8o79g40
KTfi1KZg8M3wfWYGDhSpDqdJC/1IPdEWp8VWF68zymTVa5unXRPewwQ07SP5yn19
YFlQF7l9vSnVVt4/JRPB+ydBgSXoxK6b5zbEK8+3iqBuRGvp8u0rrn4ohEkserd+
tcKssr4IEdgeVNco+UStQrrIrf+KoPN147fKzwkaUZKj3ybVExHnilr4D+HB94jL
pH404Fud+v2NWjl7RSQnsMw+gCz6Sm3eU/aWC5L5ZOecawj01Qr60nv97eqc8tdG
TrXd8yRh0cI5wL5KnO4hL/kYwOOaKsMwEkNlmL2Io7DrhVgJUAWycqfHfA==
-----END CERTIFICATE-----""".encode(
                "utf-8"
            )
        )
        tmpf.flush()
        yield tmpf.name


@fixture
def f_config_msg_signer_missing():
    with tempfile.NamedTemporaryFile() as tmpf:
        tmpf.write(
            f"""
msg_signer:
  messaging_brokers:
    - amqps://broker-01:5671
    - amqps://broker-02:5671
  messaging_cert_key: {f_client_certificate}
  messaging_ca_cert: ~/messaging/ca-cert.crt
  topic_listen_to: queue://Consumer.{{creator}}.{{task_id}}.Topic.sign.{{task_id}}
  environment: prod
  service: pubtools-sign
  timeout: 1
  retries: 3
  message_id_key: request_id
  log_level: debug""".encode(
                "utf-8"
            )
        )
        tmpf.flush()
        yield tmpf.name


@fixture
def f_environ():
    with patch.dict(os.environ, {}, clear=True) as patched:
        yield patched



from concurrent import futures
from concurrent.futures.thread import ThreadPoolExecutor
from dataclasses import dataclass, field
import datetime
import subprocess
import os
import logging
from typing import Any, Dict, List, Union, Callable, cast, Iterable

from .conf.conf import CONFIG_PATHS

from pubtools.tracing import get_trace_wrapper

tw = get_trace_wrapper()


def set_log_level(logger: logging.Logger, level: str) -> None:
    """Set log level for provided logger.

    :param logger: logger
    :type logger: logging.Logger
    :param level: logging level
    :type level: str
    """
    if level.upper() not in ("DEBUG", "INFO", "WARNING", "ERROR"):
        raise ValueError(f"Unknown log level {level}")
    logger.setLevel(level.upper())


def sanitize_log_level(log_level: str) -> str:
    """Sanitize log level. Returns INFO if provided value is invalid.

    Args:
        log_level (str): log level
    Returns:
        str: log level
    """
    if log_level.upper() not in ("DEBUG", "INFO", "WARNING", "ERROR"):
        return "INFO"
    else:
        return log_level.upper()


def isodate_now() -> str:
    """Return current datetime in ISO-8601.

    :return: str
    """
    return datetime.datetime.utcnow().isoformat() + "Z"


@tw.instrument_func(args_to_attr=True)
def run_command(cmd: List[str], env: Union[Dict[str, Any], None] = None) -> Any:
    """Run external command and return Process instance."""
    process = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, env=env
    )
    return process


def _get_config_file(config_candidate: str) -> str:
    if not os.path.exists(config_candidate):
        for config_candidate in CONFIG_PATHS:
            if os.path.exists(os.path.expanduser(config_candidate)):
                break
        else:
            raise ValueError(
                "No configuration file found: %s" % list(set(CONFIG_PATHS + [config_candidate]))
            )
    return config_candidate


@dataclass
class FData:
    """Dataclass for holding data for a function execution.

    Args:
        args (Iterable[Any]): Arguments for the function.
        kwargs (Dict[str, Any]): Keyword arguments for the function.
    """

    args: Iterable[Any]
    kwargs: Dict[str, Any] = field(default_factory=dict)


def run_in_parallel(
    func: Callable[..., Any], data: List[FData], threads: int = 10
) -> Dict[Any, Any]:
    """Run method on data in parallel.

    Args:
        func (function): Function to run on data
        data (list): List of tuples which are used as arguments for the function
    Returns:
        dict: List of result in the same order as data.
    """
    future_results = {}
    results = {}
    with ThreadPoolExecutor(max_workers=threads) as executor:
        future_results = {
            executor.submit(func, *data_entry.args, **data_entry.kwargs): n
            for n, data_entry in enumerate(data)
        }
        for future in futures.as_completed(future_results):
            if future.exception() is not None:
                raise cast(BaseException, future.exception())
            results[future_results[future]] = future.result()
    return dict(sorted(results.items(), key=lambda kv: kv[0]))



import logging

import pytest

from pubtools.sign.utils import set_log_level, sanitize_log_level, run_in_parallel, FData


def test_set_log_level():
    LOG = logging.getLogger()
    set_log_level(LOG, "DEBUG")
    assert LOG.level == logging.DEBUG
    set_log_level(LOG, "INFO")
    assert LOG.level == logging.INFO
    set_log_level(LOG, "WARNING")
    assert LOG.level == logging.WARNING
    set_log_level(LOG, "ERROR")
    assert LOG.level == logging.ERROR
    with pytest.raises(ValueError):
        set_log_level(LOG, "UNKNOWN")


def test_sanitize_log_level():
    assert sanitize_log_level("DEBUG") == "DEBUG"
    assert sanitize_log_level("INFO") == "INFO"
    assert sanitize_log_level("WARNING") == "WARNING"
    assert sanitize_log_level("Warning") == "WARNING"
    assert sanitize_log_level("ERROR") == "ERROR"
    assert sanitize_log_level("UNKNOWN") == "INFO"
    assert sanitize_log_level("unknown") == "INFO"
    assert sanitize_log_level("UnKnOwN") == "INFO"


def simulated_error(x: int) -> None:
    raise ValueError("Test")


def test_run_in_parallel_exception():
    with pytest.raises(ValueError):
        run_in_parallel(simulated_error, [FData([1])])



from typing import Any

import marshmallow as ma
from piny import MarshmallowValidator, StrictMatcher, YamlLoader

CONFIG_PATHS = ["~/.config/pubtools-sign/conf.yaml", "/etc/pubtools-sign/conf.yaml"]


class MsgSignerSchema(ma.Schema):
    """Radas signer configuration schema."""

    messaging_brokers = ma.fields.List(ma.fields.String(), required=True)
    messaging_cert_key = ma.fields.String(required=True)
    messaging_ca_cert = ma.fields.String(required=True)
    topic_send_to = ma.fields.String(required=True)
    topic_listen_to = ma.fields.String(required=True)
    environment = ma.fields.String(required=True)
    service = ma.fields.String(required=True)
    timeout = ma.fields.Integer(required=True)
    retries = ma.fields.Integer(required=True)
    send_retries = ma.fields.Integer(required=True)
    message_id_key = ma.fields.String(required=True)
    log_level = ma.fields.String(default="INFO")
    key_aliases = ma.fields.Dict(required=False, keys=ma.fields.String(), values=ma.fields.String())


class CosignSignerSchema(ma.Schema):
    """Cosign signer configuration schema."""

    cosign_bin = ma.fields.String(required=False)
    timeout = ma.fields.String(required=False)
    allow_http_registry = ma.fields.Bool(required=False)
    allow_insecure_registry = ma.fields.Bool(required=False)
    rekor_url = ma.fields.String(required=False)
    upload_tlog = ma.fields.Bool(required=False)
    log_level = ma.fields.String(default="INFO")
    env_variables = ma.fields.Dict(required=False)
    key_aliases = ma.fields.Dict(required=False, keys=ma.fields.String(), values=ma.fields.String())
    registry_user = ma.fields.String(required=False)
    registry_password = ma.fields.String(required=False)
    retries = ma.fields.Integer(required=False)


class ConfigSchema(ma.Schema):
    """pubtools-sign configuration schema."""

    msg_signer = ma.fields.Nested(MsgSignerSchema)
    cosign_signer = ma.fields.Nested(CosignSignerSchema)


def load_config(fname: str) -> Any:
    """Load configuration from a filename.

    :param fname: filename
    :type fname: str

    :return Any:
    """
    config = YamlLoader(
        path=fname,
        matcher=StrictMatcher,
        validator=MarshmallowValidator,
        schema=ConfigSchema,
    ).load(many=False)
    return config



from unittest.mock import patch

from pytest import fixture
import tempfile


@fixture
def f_msg_signer(f_config_msg_signer_ok):
    with patch("pubtools.sign.signers.msgsigner.MsgSigner") as msgsigner:
        yield msgsigner


@fixture
def f_config_msg_signer_ok(f_client_certificate, f_ca_certificate):
    with tempfile.NamedTemporaryFile() as tmpf:
        tmpf.write(
            f"""
msg_signer:
  messaging_brokers:
    - amqps://broker-01:5671
    - amqps://broker-02:5671
  messaging_cert_key: {f_client_certificate}
  messaging_ca_cert: {f_ca_certificate}
  topic_send_to: topic://Topic.sign
  topic_listen_to: queue://Consumer.{{creator}}.{{task_id}}.Topic.sign.{{task_id}}
  environment: prod
  service: pubtools-sign
  timeout: 1
  retries: 3
  send_retries: 2
  message_id_key: request_id
  log_level: debug
        """.encode(
                "utf-8"
            )
        )
        tmpf.flush()
        yield tmpf.name


@fixture
def f_config_msg_signer_ok2(f_client_certificate, f_ca_certificate, f_qpid_broker):
    qpid_broker, port = f_qpid_broker
    with tempfile.NamedTemporaryFile() as tmpf:
        tmpf.write(
            f"""
msg_signer:
  messaging_brokers:
    - localhost:{port}
  messaging_cert_key: {f_client_certificate}
  messaging_ca_cert: {f_ca_certificate}
  topic_send_to: topic://Topic.sign
  topic_listen_to: queue://Consumer.{{creator}}.{{task_id}}.Topic.sign.{{task_id}}
  environment: prod
  service: pubtools-sign
  timeout: 2
  retries: 2
  send_retries: 2
  message_id_key: request_id
  log_level: debug
        """.encode(
                "utf-8"
            )
        )
        tmpf.flush()
        yield tmpf.name


@fixture
def f_config_msg_signer_aliases(f_client_certificate):
    with tempfile.NamedTemporaryFile() as tmpf:
        tmpf.write(
            f"""
msg_signer:
  messaging_brokers:
    - amqps://broker-01:5671
    - amqps://broker-02:5671
  messaging_cert_key: {f_client_certificate}
  messaging_ca_cert: ~/messaging/ca-cert.crt
  topic_send_to: topic://Topic.sign
  topic_listen_to: queue://Consumer.{{creator}}.{{task_id}}.Topic.sign.{{task_id}}
  environment: prod
  service: pubtools-sign
  timeout: 1
  retries: 3
  send_retries: 2
  message_id_key: request_id
  log_level: debug
  key_aliases:
    beta: abcde1245
        """.encode(
                "utf-8"
            )
        )
        tmpf.flush()
        yield tmpf.name



from __future__ import annotations

import base64
from dataclasses import field, dataclass, asdict
import enum
import json
import logging
from typing import cast, Dict, List, ClassVar, Any, Optional, Type
from typing_extensions import Self
import uuid
import os
import sys

from OpenSSL import crypto
import click

from . import Signer
from ..operations.base import SignOperation
from ..operations import ClearSignOperation, ContainerSignOperation
from ..results.signing_results import SigningResults
from ..results import ClearSignResult, ContainerSignResult
from ..results import SignerResults
from ..exceptions import UnsupportedOperation
from ..clients.msg_send_client import SendClient
from ..clients.msg_recv_client import RecvClient, RecvThread
from ..models.msg import MsgMessage, MsgError
from ..conf.conf import load_config, CONFIG_PATHS
from ..utils import (
    set_log_level,
    sanitize_log_level,
    isodate_now,
    _get_config_file,
    run_in_parallel,
    FData,
)


LOG = logging.getLogger("pubtools.sign.signers.msgsigner")


class SignRequestType(str, enum.Enum):
    """Sign request type enum."""

    CONTAINER = "container_signature"
    CLEARSIGN = "clearsign_signature"


@dataclass()
class MsgSignerResults(SignerResults):
    """MsgSignerResults model."""

    status: str
    error_message: str

    def to_dict(self: SignerResults) -> Dict[Any, Any]:
        """Return dict representation of MsgSignerResults model."""
        return {"status": self.status, "error_message": self.error_message}

    @classmethod
    def doc_arguments(cls: Type[Self]) -> Dict[str, Any]:
        """Return dictionary with result description of SignerResults."""
        doc_arguments = {
            "signer_result": {
                "type": "dict",
                "description": "Signing result status.",
                "returned": "always",
                "sample": {"status": "ok", "error_message": ""},
            }
        }

        return doc_arguments


@dataclass()
class MsgSigner(Signer):
    """Messaging signer class."""

    messaging_brokers: List[str] = field(
        init=False,
        metadata={
            "description": "List of brokers URLS",
            "sample": [
                "amqps://broker-01:5671",
                "amqps://broker-02:5671",
            ],
        },
    )
    messaging_cert_key: str = field(
        init=False,
        metadata={
            "description": "Client certificate + key for messaging authorization",
            "sample": "~/messaging/cert.pem",
        },
    )
    messaging_ca_cert: str = field(
        init=False,
        metadata={"description": "Messaging CA certificate", "sample": "~/messaging/ca_cert.crt"},
    )
    topic_send_to: str = field(
        init=False,
        metadata={
            "description": "Topic where to send the messages",
            "sample": "topic://Topic.sign",
        },
    )
    topic_listen_to: str = field(
        init=False,
        metadata={
            "description": "Topic where to listen for replies",
            "sample": "queue://Consumer.{{creator}}.{{task_id}}.Topic.sign.{{task_id}}",
        },
    )
    creator: str = field(
        init=False,
        metadata={
            "description": "Identification of creator of signing request",
            "sample": "pubtools-sign",
        },
    )
    environment: str = field(
        init=False,
        metadata={"description": "Environment indetification in sent messages", "sample": "prod"},
    )
    service: str = field(
        init=False, metadata={"description": "Service identificator", "sample": "pubtools-sign"}
    )
    timeout: int = field(
        init=False,
        default=60,
        metadata={"description": "Timeout for messaging receive", "sample": 1},
    )
    retries: int = field(
        init=False,
        default=3,
        metadata={"description": "Retries for messaging receive", "sample": 3},
    )
    send_retries: int = field(
        init=False,
        default=2,
        metadata={"description": "Retries for messaging send+receive", "sample": 2},
    )
    message_id_key: str = field(
        init=False,
        metadata={
            "description": "Attribute name in message body which should be used as message id",
            "sample": "123",
        },
    )
    key_aliases: Dict[str, str] = field(
        init=False,
        metadata={
            "description": "Aliases for signing keys",
            "sample": "{'production':'abcde1245'}",
        },
        default_factory=dict,
    )

    log_level: str = field(init=False, metadata={"description": "Log level", "sample": "debug"})

    SUPPORTED_OPERATIONS: ClassVar[List[Type[SignOperation]]] = [
        ContainerSignOperation,
        ClearSignOperation,
    ]

    _signer_config_key: str = "msg_signer"

    def _construct_signing_message(
        self: MsgSigner,
        claim: str,
        signing_key: str,
        repo: str,
        extra_attrs: Optional[Dict[str, Any]] = None,
        sig_type: str = SignRequestType.CONTAINER,
    ) -> dict[str, Any]:
        data_attr = "claim_file" if sig_type == SignRequestType.CONTAINER else "data"
        _extra_attrs = extra_attrs or {}
        message = {
            "sig_key_id": signing_key[-8:],
            data_attr: claim,
            "request_id": str(uuid.uuid4()),
            "created": isodate_now(),
            "requested_by": self.creator,
            "repo": repo,
        }
        message.update(_extra_attrs)
        return message

    def _construct_headers(
        self: MsgSigner, sig_type: SignRequestType, extra_attrs: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        headers = {
            "service": self.service,
            "environment": self.environment,
            "owner_id": self.creator,
            "mtype": sig_type.value,
            "source": "metadata",
        }
        if extra_attrs:
            headers.update(extra_attrs)
        return headers

    def _create_msg_message(
        self: MsgSigner,
        data: str,
        repo: str,
        operation: SignOperation,
        sig_type: SignRequestType,
        extra_attrs: Optional[Dict[str, Any]] = None,
    ) -> MsgMessage:
        if operation.signing_key in self.key_aliases:
            signing_key = self.key_aliases[operation.signing_key]
        else:
            signing_key = operation.signing_key
        ret = MsgMessage(
            headers=self._construct_headers(sig_type, extra_attrs=extra_attrs),
            body=self._construct_signing_message(
                data,
                signing_key,
                repo,
                extra_attrs=extra_attrs,
                sig_type=sig_type.value,
            ),
            address=self.topic_send_to.format(
                **dict(list(asdict(self).items()) + list(asdict(operation).items()))
            ),
        )
        LOG.debug(f"Construted message with request_id {ret.body['request_id']}")
        return ret

    def load_config(self: MsgSigner, config_data: Dict[str, Any]) -> None:
        """Load configuration of messaging signer."""
        self.messaging_brokers = config_data["msg_signer"]["messaging_brokers"]
        self.messaging_cert_key = os.path.expanduser(
            config_data["msg_signer"]["messaging_cert_key"]
        )
        self.messaging_ca_cert = os.path.expanduser(config_data["msg_signer"]["messaging_ca_cert"])
        self.topic_send_to = config_data["msg_signer"]["topic_send_to"]
        self.topic_listen_to = config_data["msg_signer"]["topic_listen_to"]
        self.environment = config_data["msg_signer"]["environment"]
        self.service = config_data["msg_signer"]["service"]
        self.message_id_key = config_data["msg_signer"]["message_id_key"]
        self.retries = config_data["msg_signer"]["retries"]
        self.send_retries = config_data["msg_signer"]["send_retries"]
        self.log_level = config_data["msg_signer"]["log_level"]
        self.timeout = config_data["msg_signer"]["timeout"]
        self.creator = self._get_cert_subject_cn()
        self.key_aliases = config_data["msg_signer"].get("key_aliases", {})

    def _get_cert_subject_cn(self) -> str:
        x509 = crypto.load_certificate(
            crypto.FILETYPE_PEM, open(os.path.expanduser(self.messaging_cert_key)).read().encode()
        )
        return x509.get_subject().CN or x509.get_subject().UID  # type: ignore[attr-defined]

    def operations(self: MsgSigner) -> List[Type[SignOperation]]:
        """Return list of supported operations."""
        return self.SUPPORTED_OPERATIONS

    def sign(self: MsgSigner, operation: SignOperation) -> SigningResults:
        """Run signing operation.

        :param operation: signing operation
        :type operation: SignOperation

        :return: SigningResults
        """
        if isinstance(operation, ClearSignOperation):
            return self.clear_sign(operation)
        elif isinstance(operation, ContainerSignOperation):
            return self.container_sign(operation)
        else:
            raise UnsupportedOperation(operation)

    def clear_sign(self: MsgSigner, operation: ClearSignOperation) -> SigningResults:
        """Run the clearsign operation.

        :param operation: signing operation
        :type operation: ClearSignOperation

        :return: SigningResults
        """
        set_log_level(LOG, self.log_level)
        messages = []
        message_to_data = {}
        for in_data in operation.inputs:
            message = self._create_msg_message(
                base64.b64encode(in_data.encode("latin1")).decode("latin-1"),
                operation.repo,
                operation,
                SignRequestType.CLEARSIGN,
                extra_attrs={"pub_task_id": operation.task_id},
            )
            message_to_data[message.body["request_id"]] = message
            messages.append(message)

        all_messages = [x for x in messages]

        signing_key = operation.signing_key
        if signing_key in self.key_aliases:
            signing_key = self.key_aliases[signing_key]
            LOG.info(f"Using signing key alias {signing_key} for {operation.signing_key}")

        signer_results = MsgSignerResults(status="ok", error_message="")
        operation_result = ClearSignResult(
            signing_key=operation.signing_key, outputs=[""] * len(operation.inputs)
        )
        signing_results = SigningResults(
            signer=self,
            operation=operation,
            signer_results=signer_results,
            operation_result=operation_result,
        )
        errors: List[MsgError] = []
        received: Dict[int, Any] = {}
        LOG.info("errors " + str(errors))

        for i in range(self.send_retries):
            message_ids = [message.body["request_id"] for message in messages]
            LOG.debug(f"{len(messages)} messages to send")
            recvc = RecvClient(
                uid=str(i),
                message_ids=message_ids,
                topic=self.topic_listen_to.format(
                    **dict(list(asdict(self).items()) + list(asdict(operation).items()))
                ),
                id_key=self.message_id_key,
                broker_urls=self.messaging_brokers,
                cert=self.messaging_cert_key,
                ca_cert=self.messaging_ca_cert,
                timeout=self.timeout,
                retries=self.retries,
                errors=errors,
                received=received,
            )
            recvt = RecvThread(recvc)
            recvt.start()

            errors = SendClient(
                messages=messages,
                broker_urls=self.messaging_brokers,
                cert=self.messaging_cert_key,
                ca_cert=self.messaging_ca_cert,
                retries=self.retries,
                errors=errors,
            ).run()
            # check sender errors
            if errors:
                signer_results.status = "error"
                for error in errors:
                    signer_results.error_message += f"{error.name} : {error.description}\n"
                return signing_results

            # wait for receiver to finish
            recvt.join()
            recvt.stop()

            # check receiver errors
            for x in range(self.retries - 1):
                errors = recvc._errors
                if errors and errors[0].name == "MessagingTimeout":
                    LOG.info("RETRYING %s", x)
                    _messages = []
                    for message in messages:
                        if message.body["request_id"] not in received:
                            _messages.append(message)
                    if x != self.retries - 1:
                        errors.pop(0)
                    messages = _messages
                    message_ids = [message.body["request_id"] for message in messages]

                    LOG.info("Retrying recv")
                    recvc = RecvClient(
                        uid=str(i) + "-" + str(x),
                        message_ids=message_ids,
                        topic=self.topic_listen_to.format(
                            **dict(list(asdict(self).items()) + list(asdict(operation).items()))
                        ),
                        id_key=self.message_id_key,
                        broker_urls=self.messaging_brokers,
                        cert=self.messaging_cert_key,
                        ca_cert=self.messaging_ca_cert,
                        timeout=self.timeout,
                        retries=self.retries,
                        errors=errors,
                        received=received,
                    )
                    recvt = RecvThread(recvc)
                    recvt.start()
                    recvt.join()
                elif not errors:
                    break

        errors = recvc._errors
        if errors:
            signer_results.status = "error"
            for error in errors:
                signer_results.error_message += f"{error.name} : {error.description}\n"
            return signing_results

        operation_result = ClearSignResult(
            signing_key=operation.signing_key, outputs=[""] * len(all_messages)
        )

        for recv_id, _received in recvc.recv.items():
            operation_result.outputs[all_messages.index(message_to_data[recv_id])] = _received
        signing_results.operation_result = operation_result
        return signing_results

    @staticmethod
    def create_manifest_claim_message(signature_key: str, digest: str, reference: str) -> str:
        """Create manifest claim for container signing.

        See below for the specification for the manifest claim that is created here
        https://github.com/containers/image/blob/master/docs/atomic-signature.md
        """
        manifest_claim = {
            "critical": {
                "type": "atomic container signature",
                "image": {"docker-manifest-digest": digest},
                "identity": {"docker-reference": reference},
            },
            "optional": {"creator": "pubtools-sign"},
        }
        return base64.b64encode(json.dumps(manifest_claim).encode("latin1")).decode("latin1")

    def container_sign(self: MsgSigner, operation: ContainerSignOperation) -> SigningResults:
        """Run container signing operation.

        :param operation: signing operation
        :type operation: ContainerSignOperation

        :return: SigningResults
        """
        set_log_level(LOG, self.log_level)
        messages = []
        message_to_data = {}
        if len(operation.digests) != len(operation.references):
            raise ValueError("Digests must pairs with references")

        signing_key = operation.signing_key
        if signing_key in self.key_aliases:
            signing_key = self.key_aliases[signing_key]
            LOG.info(f"Using signing key alias {signing_key} for {operation.signing_key}")

        LOG.info(f"Container sign operation for {len(operation.digests)}")

        fargs = []
        for digest, reference in zip(operation.digests, operation.references):
            repo = reference.split("/", 1)[1].split(":")[0]
            fargs.append(
                FData(
                    args=[
                        self.create_manifest_claim_message(
                            signing_key, digest=digest, reference=reference
                        ),
                        repo,
                        operation,
                        SignRequestType.CONTAINER,
                    ],
                    kwargs={
                        "extra_attrs": {"pub_task_id": operation.task_id, "manifest_digest": digest}
                    },
                )
            )
        ret = run_in_parallel(self._create_msg_message, fargs)
        for n, message in ret.items():
            message_to_data[message.body["request_id"]] = message
            messages.append(message)

        all_messages = [x for x in messages]
        LOG.info(f"Signing {len(all_messages)}")

        signer_results = MsgSignerResults(status="ok", error_message="")
        operation_result = ContainerSignResult(
            signing_key=operation.signing_key, results=[""] * len(operation.digests), failed=False
        )
        signing_results = SigningResults(
            signer=self,
            operation=operation,
            signer_results=signer_results,
            operation_result=operation_result,
        )
        LOG.debug(f"{len(messages)} messages to send")

        errors: List[MsgError] = []
        received: Dict[int, Any] = {}
        LOG.info(
            "Starting signing process. Retries %d,%d, timeout: %d",
            self.send_retries,
            self.retries,
            self.timeout,
        )

        for i in range(self.send_retries):
            message_ids = [message.body["request_id"] for message in messages]
            recvc = RecvClient(
                uid=str(i),
                message_ids=message_ids,
                topic=self.topic_listen_to.format(
                    **dict(list(asdict(self).items()) + list(asdict(operation).items()))
                ),
                id_key=self.message_id_key,
                broker_urls=self.messaging_brokers,
                cert=self.messaging_cert_key,
                ca_cert=self.messaging_ca_cert,
                timeout=self.timeout,
                retries=self.retries,
                errors=errors,
                received=received,
            )
            recvt = RecvThread(recvc)
            recvt.start()

            errors = SendClient(
                messages=messages,
                broker_urls=self.messaging_brokers,
                cert=self.messaging_cert_key,
                ca_cert=self.messaging_ca_cert,
                retries=self.retries,
                errors=errors,
            ).run()

            # check sender errors
            if errors:
                signer_results.status = "error"
                for error in errors:
                    signer_results.error_message += f"{error.name} : {error.description}\n"
                return signing_results

            # wait for receiver to finish
            recvt.join()
            recvt.stop()
            received = recvc.get_received()

            for x in range(self.retries):
                errors = recvc.get_errors()
                if errors and errors[0].name == "MessagingTimeout":
                    LOG.info("Retrying receiving %s/%s", x, self.retries)
                    _messages = []
                    for message in messages:
                        if message.body["request_id"] not in received:
                            _messages.append(message)
                    if x != self.retries - 1:
                        errors.pop(0)
                    messages = _messages
                    if not messages:
                        break
                    message_ids = [message.body["request_id"] for message in messages]

                    recvc = RecvClient(
                        uid=str(i) + "-" + str(x),
                        message_ids=message_ids,
                        topic=self.topic_listen_to.format(
                            **dict(list(asdict(self).items()) + list(asdict(operation).items()))
                        ),
                        id_key=self.message_id_key,
                        broker_urls=self.messaging_brokers,
                        cert=self.messaging_cert_key,
                        ca_cert=self.messaging_ca_cert,
                        timeout=self.timeout,
                        retries=self.retries,
                        errors=errors,
                        received=received,
                    )
                    recvt = RecvThread(recvc)
                    recvt.start()
                    recvt.join()
                elif not errors:
                    break
                received = recvc.get_received()

            # check receiver errors
            errors = recvc.get_errors()
            if not errors:
                break

        if errors:
            signer_results.status = "error"
            for error in errors:
                signer_results.error_message += f"{error.name} : {error.description}\n"
            return signing_results

        operation_result = ContainerSignResult(
            signing_key=operation.signing_key, results=[""] * len(all_messages), failed=False
        )
        for recv_id, _received in recvc.recv.items():
            operation_result.failed = True if _received[0]["msg"]["errors"] else False
            operation_result.results[all_messages.index(message_to_data[recv_id])] = _received
        signing_results.operation_result = operation_result
        return signing_results


def msg_clear_sign(
    inputs: List[str],
    signing_key: str = "",
    task_id: str = "",
    config_file: str = "",
    repo: str = "",
) -> Dict[str, Any]:
    """Run clearsign operation."""
    msg_signer = MsgSigner()
    config = _get_config_file(config_file)
    msg_signer.load_config(load_config(os.path.expanduser(config)))

    str_inputs = []
    for input_ in inputs:
        if input_.startswith("@"):
            str_inputs.append(open(input_.lstrip("@")).read())
        else:
            str_inputs.append(input_)
    operation = ClearSignOperation(
        inputs=str_inputs, signing_key=signing_key, task_id=task_id, repo=repo
    )
    signing_result = msg_signer.sign(operation)
    return {
        "signer_result": signing_result.signer_results.to_dict(),
        "operation_results": cast(ClearSignResult, signing_result.operation_result).outputs,
        "operation": signing_result.operation.to_dict(),
        "signing_key": signing_result.operation_result.signing_key,
    }


def msg_container_sign(
    signing_key: str = "",
    task_id: str = "",
    config_file: str = "",
    digest: list[str] = [],
    reference: list[str] = [],
) -> Dict[str, Any]:
    """Run containersign operation with cli arguments."""
    msg_signer = MsgSigner()
    config = _get_config_file(config_file)
    msg_signer.load_config(load_config(os.path.expanduser(config)))

    operation = ContainerSignOperation(
        digests=digest,
        references=reference,
        signing_key=signing_key,
        task_id=task_id,
    )
    signing_result = msg_signer.sign(operation)
    return {
        "signer_result": signing_result.signer_results.to_dict(),
        "operation_results": signing_result.operation_result.results,  # type: ignore
        "operation": signing_result.operation.to_dict(),
        "signing_key": signing_result.operation_result.signing_key,
    }


@click.command()
@click.option(
    "--signing-key",
    required=True,
    help="8 characters key fingerprint of key which should be used for signing",
)
@click.option("--task-id", required=True, help="Task id identifier (usually pub task-id)")
@click.option("--config-file", default=CONFIG_PATHS[0], help="path to the config file")
@click.option("--raw", default=False, is_flag=True, help="Print raw output instead of json")
@click.option(
    "--log-level",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR"]),
    default="INFO",
    help="Set log level",
)
@click.option("--repo", help="Repository reference")
@click.argument("inputs", nargs=-1)
def msg_clear_sign_main(
    inputs: List[str],
    signing_key: str = "",
    task_id: str = "",
    config_file: str = "",
    raw: bool = False,
    log_level: str = "INFO",
    repo: str = "",
) -> None:
    """Entry point method for clearsign operation."""
    ch = logging.StreamHandler()
    ch.setLevel(getattr(logging, sanitize_log_level(log_level)))

    LOG.addHandler(ch)
    logging.basicConfig(level=getattr(logging, sanitize_log_level(log_level)))

    ret = msg_clear_sign(
        inputs,
        signing_key=signing_key,
        task_id=task_id,
        repo=repo,
        config_file=config_file,
    )
    if not raw:
        click.echo(json.dumps(ret))
        if ret["signer_result"]["status"] == "error":
            sys.exit(1)
    else:
        if ret["signer_result"]["status"] == "error":
            print(ret["signer_result"]["error_message"], file=sys.stderr)
            sys.exit(1)
        else:
            for claim in ret["operation_results"]:
                if claim[0]["msg"]["errors"]:
                    for error in claim[0]["msg"]["errors"]:
                        print(error, file=sys.stderr)
                    sys.exit(1)
                else:
                    print(claim[0]["msg"]["signed_data"])


@click.command()
@click.option(
    "--signing-key",
    required=True,
    help="8 characters key fingerprint of key which should be used for signing",
)
@click.option("--task-id", required=True, help="Task id identifier (usually pub task-id)")
@click.option("--config-file", default=CONFIG_PATHS[0], help="path to the config file")
@click.option(
    "--digest",
    required=True,
    multiple=True,
    type=str,
    help="Digests which should be signed.",
)
@click.option(
    "--reference",
    required=True,
    multiple=True,
    type=str,
    help="References which should be signed.",
)
@click.option("--raw", default=False, is_flag=True, help="Print raw output instead of json")
@click.option(
    "--log-level",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR"]),
    default="INFO",
    help="Set log level",
)
def msg_container_sign_main(
    signing_key: str = "",
    task_id: str = "",
    config_file: str = "",
    digest: List[str] = [],
    reference: List[str] = [],
    raw: bool = False,
    log_level: str = "INFO",
) -> None:
    """Entry point method for containersign operation."""
    ch = logging.StreamHandler()
    ch.setLevel(getattr(logging, sanitize_log_level(log_level)))
    LOG.addHandler(ch)
    logging.basicConfig(level=getattr(logging, sanitize_log_level(log_level)))

    ret = msg_container_sign(
        signing_key=signing_key,
        task_id=task_id,
        config_file=config_file,
        digest=digest,
        reference=reference,
    )
    if not raw:
        click.echo(json.dumps(ret))
        if ret["signer_result"]["status"] == "error":
            sys.exit(1)
    else:
        for claim in ret["operation_results"]:
            if claim[0]["msg"]["errors"]:
                for error in claim[0]["msg"]["errors"]:
                    print(error, file=sys.stderr)
                sys.exit(1)
            else:
                print(claim[0]["msg"]["signed_claim"])



# flake8: noqa: F401

from __future__ import annotations

from .clearsign import ClearSignOperation
from .containersign import ContainerSignOperation

__all__ = ["ClearSignOperation", "ContainerSignOperation"]



from __future__ import annotations

from abc import ABC, abstractmethod
import dataclasses
from typing import Any, Dict

from .clearsign import ClearSignResult  # noqa: F401
from .containersign import ContainerSignResult  # noqa: F401

__all__ = ["ClearSignResult", "ContainerSignResult"]


@dataclasses.dataclass()
class SignerResults(ABC):
    """SignerResults abstract class."""

    status: str
    error_message: str

    @abstractmethod
    def to_dict(self: SignerResults) -> Dict[Any, Any]:
        """Return dict representation of SignerResults."""
        ...  # pragma: no cover



import dataclasses
import base64
import json
import os
import logging

from urllib.parse import urlparse, urlunparse, urlencode
from urllib.request import parse_http_list, parse_keqv_list
from typing import Optional, Tuple, Any, Union

import requests
from requests.adapters import HTTPAdapter, Retry

from ..utils import set_log_level

LOG = logging.getLogger("pubtools.sign.signers.msgsigner")

AUTH_FILES = [
    "${XDG_CONFIG_HOME}/containers/auth.json",
    "${HOME}/.docker/config.json",
    "${REGISTRY_AUTH_FILE}",
]


@dataclasses.dataclass
class AuthTokenWrapper:
    """Carrier of the auth token for container registry."""

    token: str


class ContainerRegistryClient:
    """Client for interacting with container registries."""

    def __init__(
        self,
        username: Optional[str] = None,
        password: Optional[str] = None,
        auth_file: Optional[str] = None,
        retries: int = 5,
        log_level: str = "INFO",
    ):
        """Initialize.

        Args:
            username (Optional[str]): Username for authentication.
            password (Optional[str]): Password for authentication.
            auth_file (Optional[str]): Path to the auth file.
            retries (int): Number of retries for HTTP requests.
        """
        self.username = username
        self.password = password
        self.auth_file = auth_file
        self._session: Union[None, requests.Session] = None
        self.retries = retries
        set_log_level(LOG, log_level)

    @property
    def session(self) -> requests.Session:
        """Get the session object."""
        if not self._session:
            self._session = requests.Session()
            retries = Retry(
                total=self.retries, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504]
            )
            self._session.mount("http://", HTTPAdapter(max_retries=retries))
        return self._session

    def resolve_authentication(self, image_reference: str) -> Tuple[str, str]:
        """Resolve authentication for the given image reference.

        When username and password are provided in registry client, they are used.
        Otherwise container configuration files are search for specific authentication
        entry based on host of image reference.

        Args:
            image_reference (str): Image reference to resolve authentication for.
        Returns:
            Tuple[str, str]: Username and password for authentication.
        """
        if self.username and self.password:
            return (self.username, self.password)

        parsed = urlparse(image_reference)
        if not parsed.scheme:
            parsed = urlparse(f"docker://{image_reference}")
        registry = parsed.netloc
        existing_auth_files = []
        auth_files = AUTH_FILES if not self.auth_file else [self.auth_file] + AUTH_FILES
        for af in auth_files:
            if os.path.exists(os.path.expandvars(af)):
                existing_auth_files.append(os.path.expandvars(af))
        for eaf in existing_auth_files:
            parsed_conf = json.load(open(os.path.expandvars(eaf)))
            parsed_auths = parsed_conf.get("auths")
            if registry in parsed_auths:
                auth = (
                    base64.b64decode(parsed_auths.get(registry)["auth"].encode("utf-8"))
                    .decode("utf-8")
                    .split(":")
                )
                auth_tuple = (auth[0], auth[1])
                break
        else:
            raise ValueError("No authentication found")
        return auth_tuple

    def authenticate_to_registry(self, image_reference: str, auth_header: str) -> Union[str, Any]:
        """Ask for auth token based on given auth header.

        Args:
            image_reference (str): Image reference to resolve authentication for.
            auth_header (str): Authentication header from the registry.
        Returns:
            str: Authentication token.
        """
        _, _, value = auth_header.partition("Bearer")
        items = parse_http_list(value)
        opts = parse_keqv_list(items)
        unparse_parts = [
            "https",
            opts["realm"].replace("https://", "").split("/", 1)[0],
            opts["realm"].replace("https://", "").split("/", 1)[1],
            "",
            urlencode({"service": opts["service"], "scope": opts["scope"]}),
            "",
        ]
        auth_url = urlunparse(unparse_parts)
        username, password = self.resolve_authentication(image_reference)
        response = self.session.get(auth_url, auth=(username, password))

        response.raise_for_status()
        return response.json().get("token")

    def check_container_image_exists(
        self, image_reference: str, auth_token: AuthTokenWrapper
    ) -> Tuple[bool, str]:
        """Check if the given container image exists.

        Args:
            image_reference (str): Image reference to check.
            auth_token (AuthTokenWrapper): Authentication token.
        Returns:
            bool: [True, ""] if the image exists, Tuple[False, <error_message>] otherwise.
        """
        repo_ref, tag = image_reference.rsplit(":", 1)
        registry, repo = repo_ref.split("/", 1)
        manifest_url = f"https://{registry}/v2/{repo}/manifests/{tag}"
        headers = {"Authorization": f"Bearer {auth_token.token}"}
        response = self.session.get(manifest_url, headers=headers)

        if response.status_code == 200:
            return True, ""
        elif response.status_code == 401:
            auth_header = response.headers["www-authenticate"]
            auth_token.token = self.authenticate_to_registry(image_reference, auth_header)
            # Retry the original request with the token
            headers = {"Authorization": f"Bearer {auth_token.token}"}
            response = self.session.get(manifest_url, headers=headers)
            if response.status_code == 200:
                return True, ""
            elif response.status_code == 404:
                return False, ""
            else:
                LOG.error(f"Unexpected Error: {response.status_code} - {response.text}")
                return False, f"Unexpected Error: {response.status_code} - {response.text}"
        elif response.status_code == 404:
            return False, ""
        else:
            LOG.error(f"Unexpected Error: {response.status_code} - {response.text}")
            return False, f"Unexpected Error: {response.status_code} - {response.text}"



import datetime
import json
import logging
import threading
from typing import Any, List, Dict, Union

from ..models.msg import MsgError

from .msg import _MsgClient

import proton
import proton.utils
from proton.reactor import Container

from pubtools.tracing import get_trace_wrapper

tw = get_trace_wrapper()
LOG = logging.getLogger("pubtools.sign.client.msg_recv_client")


class _RecvClient(_MsgClient):
    def __init__(
        self,
        uid: str,
        topic: str,
        message_ids: List[str],
        id_key: str,
        broker_urls: List[str],
        cert: str,
        ca_cert: str,
        timeout: int,
        recv: Dict[Any, Any],
        errors: List[MsgError],
    ) -> None:
        super().__init__(errors=errors)
        self.broker_urls = broker_urls
        self.topic = topic
        self.id_key = id_key
        self.ssl_domain = proton.SSLDomain(proton.SSLDomain.MODE_CLIENT)
        if cert:
            self.ssl_domain.set_credentials(cert, cert, None)
        if ca_cert:
            self.ssl_domain.set_trusted_ca_db(ca_cert)
        self.ssl_domain.set_peer_authentication(proton.SSLDomain.ANONYMOUS_PEER)
        self.recv_ids = {x: False for x in message_ids}
        self.confirmed = 0
        self.recv = recv
        self.timeout = timeout
        self.recv_in_time = False
        self.uid = uid
        self.last_message_received = datetime.datetime.now()
        LOG.info("Expected to receive %s messages", len(message_ids))

    def on_start(self, event: proton.Event) -> None:
        LOG.debug("RECEIVER: On start %s %s %s", event, self.topic, self.broker_urls)
        self.conn = event.container.connect(
            urls=self.broker_urls, ssl_domain=self.ssl_domain, sasl_enabled=False
        )
        self.receiver = event.container.create_receiver(self.conn, self.topic)
        self.timer_task = event.container.schedule(self.timeout / 2, self)

    @tw.instrument_func()
    def on_message(self, event: proton.Event) -> None:
        LOG.debug("RECEIVER: On message (%s)", event)
        outer_message = json.loads(event.message.body)
        headers = event.message.properties
        msg_id = outer_message["msg"][self.id_key]

        if msg_id in self.recv_ids:
            self.recv_ids[msg_id] = True
            self.recv[msg_id] = (outer_message, headers)
            self.recv_in_time = True
            self.last_message_received = datetime.datetime.now()
            self.accept(event.delivery)
        else:
            LOG.debug(f"RECEIVER: Ignored message {msg_id}")

        if self.recv_ids.values() and all(self.recv_ids.values()):
            self.timer_task.cancel()
            event.receiver.close()
            event.connection.close()
            LOG.info("[%d][%s] All messages received", threading.get_ident(), self.uid)

    def on_timer_task(self, event: proton.Event) -> None:
        if self.recv_in_time:
            LOG.info(
                "[%d][%s] RECEIVER: On timeout but messages was received "
                "- continue, received: %d/%d",
                threading.get_ident(),
                self.uid,
                len([x for x in self.recv_ids.values() if x]),
                len(self.recv_ids),
            )
            self.recv_in_time = False
            self.timer_task = event.reactor.schedule(self.timeout / 2, self)
            return
        if (datetime.datetime.now() - self.last_message_received).total_seconds() < self.timeout:
            self.timer_task = event.reactor.schedule(self.timeout / 2, self)
            return
        LOG.info(
            "[%d][%s] RECEIVER: On timeout (%s) messages: %d/%d",
            threading.get_ident(),
            event,
            self.uid,
            len([x for x in self.recv_ids.values() if x]),
            len(self.recv_ids),
        )
        self.timer_task.cancel()
        if event.connection:
            event.connection.close()  # pragma: no cover
        if event.receiver:
            event.receiver.close()  # pragma: no cover
        event.container.stop()

        if not all(self.recv_ids.values()):
            self.errors.append(
                MsgError(
                    source=event,
                    name="MessagingTimeout",
                    description="[%d] Out of time when receiving messages (%d/%d)"
                    % (
                        threading.get_ident(),
                        len([x for x in self.recv_ids.values() if x]),
                        len(self.recv_ids),
                    ),
                )
            )

    def close(self) -> None:
        if hasattr(self, "timer_task"):
            self.timer_task.cancel()
        if hasattr(self, "receiver"):
            self.receiver.close()
        if hasattr(self, "conn"):
            self.conn.close()


class RecvClient(Container):
    """Messaging receiver."""

    def __init__(
        self,
        uid: str,
        topic: str,
        message_ids: List[str],
        id_key: str,
        broker_urls: List[str],
        cert: str,
        ca_cert: str,
        timeout: int,
        retries: int,
        errors: List[MsgError],
        received: Dict[Any, Any],
    ) -> None:
        """Recv Client Initializer.

        :param topic: Topic where to listen for incoming messages (for example topic://Topic.signed)
        :type topic: str
        :param message_ids: List of awaited message ids
        :type topic: List[str]
        :param id_key: Attribute name in message body which is considered as id
        :type topic: str
        :param message_ids: List of broker urls
        :type topic: List[str]
        :param cert: Messaging client certificate
        :type cert: str
        :param ca_cert: Messaging ca certificate
        :type ca_cert: str
        :param timeout: Timeout for the messaging receiver
        :type timeout: int
        :param retries: How many attempts to retry receiving messages
        :type retries: int
        :param errors: List of errors which occured during the process
        :type errors: List[MsgError]
        :param received: Mapping of received messages
        :type errors: Dict[int, Any]
        """
        self.message_ids = message_ids
        self.recv: Dict[Any, Any] = received
        self._errors: List[MsgError] = errors
        self.topic = topic
        self.message_ids = message_ids
        self.id_key = id_key
        self.broker_urls = broker_urls
        self.cert = cert
        self.ca_cert = ca_cert
        self.timeout = timeout
        self.uid = uid
        self._retries = retries
        handler = _RecvClient(
            uid=uid,
            topic=topic,
            message_ids=message_ids,
            id_key=id_key,
            broker_urls=broker_urls,
            cert=cert,
            ca_cert=ca_cert,
            timeout=timeout,
            recv=self.recv,
            errors=self._errors,
        )
        super().__init__(handler)
        self._handler = handler

    def get_errors(self) -> List[MsgError]:
        """Get errors from receiver.

        This method doesn't have any meaningfull usecase, it's only used for testing
        """
        return self._errors  # pragma: no cover

    def get_received(self) -> Dict[Any, Any]:
        """Get received messages.

        This method doesn't have any meaningfull usecase, it's only used for testing
        """
        return self.recv  # pragma: no cover

    def run(self) -> Union[Dict[Any, Any], List[MsgError]]:
        """Run the receiver."""
        if not len(self.message_ids):
            LOG.warning("No messages to receive")
            return []
        super().run()
        if self._errors:
            return self._errors
        return self.recv

    def close(self) -> None:
        """Close receiver."""
        if self._handler:
            self._handler.close()


class RecvThread(threading.Thread):
    """Receiver wrapper allows to stop receiver on demand."""

    def __init__(self, recv: RecvClient):
        """Recv Thread Initializer.

        Args:
            recv (RecvClient): RecvClient instance
        """
        super().__init__()
        self.recv = recv

    def stop(self) -> None:
        """Stop receiver."""
        self.recv.close()

    def run(self) -> None:
        """Run receiver."""
        self.recv.run()
